//
// Version resource script.
//

#include <winver.h>
#include <itcl.h>

//
// build-up the name suffix that defines the type of build this is.
//
#if DEBUG && !UNCHECKED
#define SUFFIX_DEBUG	    "g"
#else
#define SUFFIX_DEBUG	    ""
#endif

#define SUFFIX		    SUFFIX_DEBUG


VS_VERSION_INFO VERSIONINFO
 FILEVERSION	ITCL_MAJOR_VERSION,ITCL_MINOR_VERSION,ITCL_RELEASE_LEVEL,ITCL_RELEASE_SERIAL
 PRODUCTVERSION	ITCL_MAJOR_VERSION,ITCL_MINOR_VERSION,ITCL_RELEASE_LEVEL,ITCL_RELEASE_SERIAL
 FILEFLAGSMASK	0x3fL
#ifdef DEBUG
 FILEFLAGS 	VS_FF_DEBUG
#else
 FILEFLAGS 	0x0L
#endif
 FILEOS 	VOS__WINDOWS32
 FILETYPE 	VFT_DLL
 FILESUBTYPE	0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
	BLOCK "040904b0"
	BEGIN
	    VALUE "FileDescription", "Itcl language extension for Tcl\0"
	    VALUE "Authors", "Michael McLennan\0"
	    VALUE "OriginalFilename", "itcl" STRINGIFY(ITCL_MAJOR_VERSION) STRINGIFY(ITCL_MINOR_VERSION) SUFFIX ".dll\0"
	    VALUE "CompanyName", "Bell Labs Innovations for Lucent Technologies\0"
	    VALUE "FileVersion", ITCL_PATCH_LEVEL
	    VALUE "LegalCopyright", "Copyright \251 1993-2004\0"
	    VALUE "ProductName", "[Incr Tcl] " ITCL_VERSION " for Windows\0"
	    VALUE "ProductVersion", ITCL_PATCH_LEVEL
	END
    END
    BLOCK "VarFileInfo"
    BEGIN
	VALUE "Translation", 0x409, 1200
    END
END




