/* 
 * itclStubInit.c --
 *
 *	This file contains the initializers for the Itcl stub vectors.
 *
 * Copyright (c) 1998-1999 by XXX
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "itclInt.h"

/*
 * Remove macros that will interfere with the definitions below.
 */


/*
 * WARNING: The contents of this file is automatically generated by the
 * tools/genStubs.tcl script. Any modifications to the function declarations
 * below should be made in the generic/tcl.decls script.
 */

/* !BEGIN!: Do not edit below this line. */

ItclIntStubs itclIntStubs = {
    TCL_STUB_MAGIC,
    NULL,
    Itcl_IsClassNamespace, /* 0 */
    Itcl_IsClass, /* 1 */
    Itcl_FindClass, /* 2 */
    Itcl_FindObject, /* 3 */
    Itcl_IsObject, /* 4 */
    Itcl_ObjectIsa, /* 5 */
    Itcl_Protection, /* 6 */
    Itcl_ProtectionStr, /* 7 */
    Itcl_CanAccess, /* 8 */
    Itcl_CanAccessFunc, /* 9 */
    Itcl_GetTrueNamespace, /* 10 */
    Itcl_ParseNamespPath, /* 11 */
    Itcl_DecodeScopedCommand, /* 12 */
    Itcl_EvalArgs, /* 13 */
    Itcl_CreateArgs, /* 14 */
    Itcl_PushContext, /* 15 */
    Itcl_PopContext, /* 16 */
    Itcl_GetContext, /* 17 */
    Itcl_InitHierIter, /* 18 */
    Itcl_DeleteHierIter, /* 19 */
    Itcl_AdvanceHierIter, /* 20 */
    Itcl_FindClassesCmd, /* 21 */
    Itcl_FindObjectsCmd, /* 22 */
    Itcl_ProtectionCmd, /* 23 */
    Itcl_DelClassCmd, /* 24 */
    Itcl_DelObjectCmd, /* 25 */
    Itcl_ScopeCmd, /* 26 */
    Itcl_CodeCmd, /* 27 */
    Itcl_StubCreateCmd, /* 28 */
    Itcl_StubExistsCmd, /* 29 */
    Itcl_IsStub, /* 30 */
    Itcl_CreateClass, /* 31 */
    Itcl_DeleteClass, /* 32 */
    Itcl_FindClassNamespace, /* 33 */
    Itcl_HandleClass, /* 34 */
    Itcl_ClassCmdResolver, /* 35 */
    Itcl_ClassVarResolver, /* 36 */
    Itcl_ClassCompiledVarResolver, /* 37 */
    Itcl_BuildVirtualTables, /* 38 */
    Itcl_CreateVarDefn, /* 39 */
    Itcl_DeleteVarDefn, /* 40 */
    Itcl_GetCommonVar, /* 41 */
    Itcl_CreateMember, /* 42 */
    Itcl_DeleteMember, /* 43 */
    Itcl_CreateObject, /* 44 */
    Itcl_DeleteObject, /* 45 */
    Itcl_DestructObject, /* 46 */
    Itcl_HandleInstance, /* 47 */
    Itcl_GetInstanceVar, /* 48 */
    Itcl_ScopedVarResolver, /* 49 */
    Itcl_BodyCmd, /* 50 */
    Itcl_ConfigBodyCmd, /* 51 */
    Itcl_CreateMethod, /* 52 */
    Itcl_CreateProc, /* 53 */
    Itcl_CreateMemberFunc, /* 54 */
    Itcl_ChangeMemberFunc, /* 55 */
    Itcl_DeleteMemberFunc, /* 56 */
    Itcl_CreateMemberCode, /* 57 */
    Itcl_DeleteMemberCode, /* 58 */
    Itcl_GetMemberCode, /* 59 */
    NULL, /* 60 */
    Itcl_EvalMemberCode, /* 61 */
    Itcl_CreateArgList, /* 62 */
    Itcl_CreateArg, /* 63 */
    Itcl_DeleteArgList, /* 64 */
    Itcl_ArgList, /* 65 */
    Itcl_EquivArgLists, /* 66 */
    Itcl_GetMemberFuncUsage, /* 67 */
    Itcl_ExecMethod, /* 68 */
    Itcl_ExecProc, /* 69 */
    Itcl_AssignArgs, /* 70 */
    Itcl_ConstructBase, /* 71 */
    Itcl_InvokeMethodIfExists, /* 72 */
    NULL, /* 73 */
    Itcl_ReportFuncErrors, /* 74 */
    Itcl_ParseInit, /* 75 */
    Itcl_ClassCmd, /* 76 */
    Itcl_ClassInheritCmd, /* 77 */
    Itcl_ClassProtectionCmd, /* 78 */
    Itcl_ClassConstructorCmd, /* 79 */
    Itcl_ClassDestructorCmd, /* 80 */
    Itcl_ClassMethodCmd, /* 81 */
    Itcl_ClassProcCmd, /* 82 */
    Itcl_ClassVariableCmd, /* 83 */
    Itcl_ClassCommonCmd, /* 84 */
    Itcl_ParseVarResolver, /* 85 */
    Itcl_BiInit, /* 86 */
    Itcl_InstallBiMethods, /* 87 */
    Itcl_BiIsaCmd, /* 88 */
    Itcl_BiConfigureCmd, /* 89 */
    Itcl_BiCgetCmd, /* 90 */
    Itcl_BiChainCmd, /* 91 */
    Itcl_BiInfoClassCmd, /* 92 */
    Itcl_BiInfoInheritCmd, /* 93 */
    Itcl_BiInfoHeritageCmd, /* 94 */
    Itcl_BiInfoFunctionCmd, /* 95 */
    Itcl_BiInfoVariableCmd, /* 96 */
    Itcl_BiInfoBodyCmd, /* 97 */
    Itcl_BiInfoArgsCmd, /* 98 */
    Itcl_DefaultInfoCmd, /* 99 */
    Itcl_EnsembleInit, /* 100 */
    Itcl_CreateEnsemble, /* 101 */
    Itcl_AddEnsemblePart, /* 102 */
    Itcl_GetEnsemblePart, /* 103 */
    Itcl_IsEnsemble, /* 104 */
    Itcl_GetEnsembleUsage, /* 105 */
    Itcl_GetEnsembleUsageForObj, /* 106 */
    Itcl_EnsembleCmd, /* 107 */
    Itcl_EnsPartCmd, /* 108 */
    Itcl_EnsembleErrorCmd, /* 109 */
    NULL, /* 110 */
    NULL, /* 111 */
    _Tcl_GetCallFrame, /* 112 */
    _Tcl_ActivateCallFrame, /* 113 */
    _TclNewVar, /* 114 */
    Itcl_Assert, /* 115 */
    Itcl_IsObjectCmd, /* 116 */
    Itcl_IsClassCmd, /* 117 */
};

static ItclStubHooks itclStubHooks = {
    &itclIntStubs
};

ItclStubs itclStubs = {
    TCL_STUB_MAGIC,
    &itclStubHooks,
    Itcl_Init, /* 0 */
    Itcl_SafeInit, /* 1 */
    Itcl_RegisterC, /* 2 */
    Itcl_RegisterObjC, /* 3 */
    Itcl_FindC, /* 4 */
    Itcl_InitStack, /* 5 */
    Itcl_DeleteStack, /* 6 */
    Itcl_PushStack, /* 7 */
    Itcl_PopStack, /* 8 */
    Itcl_PeekStack, /* 9 */
    Itcl_GetStackValue, /* 10 */
    Itcl_InitList, /* 11 */
    Itcl_DeleteList, /* 12 */
    Itcl_CreateListElem, /* 13 */
    Itcl_DeleteListElem, /* 14 */
    Itcl_InsertList, /* 15 */
    Itcl_InsertListElem, /* 16 */
    Itcl_AppendList, /* 17 */
    Itcl_AppendListElem, /* 18 */
    Itcl_SetListValue, /* 19 */
    Itcl_EventuallyFree, /* 20 */
    Itcl_PreserveData, /* 21 */
    Itcl_ReleaseData, /* 22 */
    Itcl_SaveInterpState, /* 23 */
    Itcl_RestoreInterpState, /* 24 */
    Itcl_DiscardInterpState, /* 25 */
};

/* !END!: Do not edit above this line. */
